from __future__ import annotations

from dataclasses import dataclass

from .environment import EnvironmentManager
from .network import Network
from .os_info import OSInfo
from .python_info import PythonInfo
from .resource_limits import ResourceLimits


@dataclass
class ExecutionContext:
    os_info: OSInfo
    network: Network
    python_info: PythonInfo
    environment: EnvironmentManager
    resource_limits: ResourceLimits

    @classmethod
    def create(cls) -> ExecutionContext:
        return cls(
            environment=EnvironmentManager.load(),
            network=Network.load(),
            python_info=PythonInfo.load(),
            os_info=OSInfo.load(),
            resource_limits=ResourceLimits.load(),
        )



import os
import signal
import types
from typing import Any, Callable


SIGNAL_TRANSLATION_MAP: dict[int, str] = {
    signal.SIGINT: "SIGINT",
    signal.SIGTERM: "SIGTERM",
}


class DelayedInterrupt:
    _pid: int
    _propagate_to_forked_processes: bool | None
    _sig: int | None
    _frame: types.FrameType | None
    _old_signal_handler_map: dict[
        int, Callable[[int, types.FrameType | None], None] | int | None
    ]

    def __init__(self, propagate_to_forked_processes: bool | None = None):
        self._pid = os.getpid()
        self._propagate_to_forked_processes = propagate_to_forked_processes
        self._sig = None
        self._frame = None
        self._old_signal_handler_map = {}

    def __enter__(self) -> None:
        self._old_signal_handler_map = {
            sig: signal.signal(sig, self._handler)
            for sig, _ in SIGNAL_TRANSLATION_MAP.items()
        }

    def __exit__(self, *args: Any, **kwargs: Any) -> None:
        for sig, handler in self._old_signal_handler_map.items():
            signal.signal(sig, handler)

        if self._sig is None:
            return

        signal_handler = self._old_signal_handler_map[self._sig]
        if callable(signal_handler):
            signal_handler(self._sig, self._frame)

    def _handler(self, sig: int, frame: types.FrameType | None) -> None:
        self._sig = sig
        self._frame = frame

        if os.getpid() != self._pid:
            if self._propagate_to_forked_processes is False:
                print(
                    "!!! DelayedKeyboardInterrupt._handler: "
                    f"{SIGNAL_TRANSLATION_MAP[sig]} received; "
                    f"PID mismatch: {os.getpid()=}, {self._pid=}, "
                    "calling original handler"
                )
                signal_handler = self._old_signal_handler_map[self._sig]
                if callable(signal_handler):
                    signal_handler(self._sig, self._frame)
            elif self._propagate_to_forked_processes is None:
                print(
                    "!!! DelayedKeyboardInterrupt._handler: "
                    f"{SIGNAL_TRANSLATION_MAP[sig]} received; "
                    f"PID mismatch: {os.getpid()=}, ignoring the signal"
                )
                return

        print(
            "!!! DelayedKeyboardInterrupt._handler: "
            f"{SIGNAL_TRANSLATION_MAP[sig]} received; "
            "delaying KeyboardInterrupt"
        )



import asyncio
from typing import Type

from .execution_context import ExecutionContext
from .imicroservice import IMicroService
from .lifespan import DelayedInterrupt


class BaseLauncher:
    microservice: Type[IMicroService]

    def __init__(self, microservice: Type[IMicroService]) -> None:
        self.microservice = microservice

    def run(
        self, loop: asyncio.AbstractEventLoop = asyncio.new_event_loop()
    ) -> None:
        instance = self.microservice(ExecutionContext.create())
        self.main(instance, loop)

    def main(
        self, instance: IMicroService, loop: asyncio.AbstractEventLoop
    ) -> None:
        try:
            self.setup(instance, loop)
            loop.run_until_complete(instance.main())
        finally:
            self.shutdown(instance, loop)

    def setup(
        self, instance: IMicroService, loop: asyncio.AbstractEventLoop
    ) -> None:
        with DelayedInterrupt():
            loop.run_until_complete(instance.setup())

    def shutdown(
        self, instance: IMicroService, loop: asyncio.AbstractEventLoop
    ) -> None:
        with DelayedInterrupt():
            loop.run_until_complete(instance.shutdown())
            try:
                self._cancel_all_task(loop)
                loop.run_until_complete(loop.shutdown_asyncgens())
            finally:
                loop.close()

    def _cancel_all_task(self, loop: asyncio.AbstractEventLoop) -> None:
        to_cancel = asyncio.tasks.all_tasks(loop)
        if not to_cancel:
            return

        for task in to_cancel:
            task.cancel()

        loop.run_until_complete(
            asyncio.tasks.gather(*to_cancel, return_exceptions=True)
        )

        for task in filter(lambda task: not task.cancelled(), to_cancel):
            if task_exc := task.exception():
                loop.call_exception_handler(
                    {
                        "message": "unhandled exception during shutdown",
                        "exception": task_exc,
                        "task": task,
                    }
                )



from __future__ import annotations

import socket
from dataclasses import dataclass

from .utils import parse_dns_servers_from_file


@dataclass(frozen=True)
class Network:
    hostname: str | None
    dns_servers: list[str]
    ip_addresses: list[str]
    alternative_hostnames: list[str]
    default_socket_timeout: float | None
    network_interfaces: list[tuple[int, str]]

    @classmethod
    def load(cls) -> Network:
        hostname = socket.gethostname()
        ip_address = socket.gethostbyname(hostname)
        alternative_hostnames: list[str] = []
        list_of_ips: list[str] = []
        try:
            (
                _,
                alternative_hostnames,
                list_of_ips,
            ) = socket.gethostbyaddr(ip_address)
        except socket.herror:
            pass
        network_interfaces = socket.if_nameindex()
        dns_servers = parse_dns_servers_from_file()
        return cls(
            hostname=hostname,
            alternative_hostnames=alternative_hostnames,
            ip_addresses=list_of_ips,
            network_interfaces=network_interfaces,
            dns_servers=dns_servers,
            default_socket_timeout=socket.getdefaulttimeout(),
        )



from .environment import (
    DupOperationException,
    EnvConfValue,
    EnvironmentConfig,
    EnvironmentException,
    EnvironmentManager,
    ImmutableDict,
    LazyValue,
    LazyValueException,
    NoValueException,
)
from .execution_context import ExecutionContext
from .imicroservice import IMicroService
from .launcher import BaseLauncher
from .network import Network
from .os_info import OSInfo
from .python_info import PythonInfo
from .resource_limits import Limit, ResourceLimits
from .utils import get_resource_limit, parse_dns_servers_from_file

__all__ = (
    "IMicroService",
    "BaseLauncher",
    "Network",
    "parse_dns_servers_from_file",
    "PythonInfo",
    "OSInfo",
    "get_resource_limit",
    "ResourceLimits",
    "Limit",
    "ExecutionContext",
    "EnvironmentManager",
    "ImmutableDict",
    "EnvironmentException",
    "EnvConfValue",
    "EnvironmentConfig",
    "NoValueException",
    "LazyValueException",
    "DupOperationException",
    "LazyValue",
)



from __future__ import annotations

from typing import Any, Callable, Generic, TypeVar

from .manager import EnvironmentManager


T = TypeVar("T")


class EnvConfValue(Generic[T]):
    key: str
    optional: bool
    default: str | None
    _loaded: bool
    _value: T | None
    validator: Callable[[str], None] | None
    converter: Callable[[str], Any] | None

    def __init__(
        self,
        key: str,
        default: str | None = None,
        optional: bool = False,
        validator: Callable[[str], None] | None = None,
        converter: Callable[[str], Any] | None = None,
    ) -> None:
        self.key = key
        self.default = default
        self.optional = optional
        self.validator = validator
        self.converter = converter
        self._loaded = False
        self._value = None

    def load_value(self, environment: EnvironmentManager) -> None:
        loader = environment.get(self.key)
        if self.optional:
            loader.optional()
        if self.validator:
            loader.validator(self.validator)
        if self.converter:
            loader.converter(self.converter)
        if self.default:
            loader.default(self.default)
        self._value = loader.resolve()
        self._loaded = True

    @property
    def value(self) -> T | None:
        if not self._loaded:
            raise ValueError(f'EnvConfValue "{self.key}" not loaded value yet')
        return self._value


class EnvironmentConfig:
    def __init__(self, environment: EnvironmentManager) -> None:
        for value in filter(
            lambda item: isinstance(item, EnvConfValue),
            self.__class__.__dict__.values(),
        ):
            value.load_value(environment)



class EnvironmentException(Exception):
    pass


class NoValueException(EnvironmentException):
    pass


class LazyValueException(EnvironmentException):
    pass


class DupOperationException(LazyValueException):
    pass



from __future__ import annotations

from typing import Any, Callable

from .exceptions import DupOperationException, NoValueException
from .imanager import IEnvironmentManager


class LazyValue:
    key: str
    is_optional: bool
    default_value: str | None
    converter_func: Callable[[str], Any] | None
    validator_func: Callable[[str], None] | None
    environment: IEnvironmentManager

    def __init__(self, key: str, environment: IEnvironmentManager) -> None:
        self.key = key
        self.environment = environment
        self.is_optional = False
        self.default_value = None
        self.converter_func = None
        self.validator_func = None

    def default(self, value: str) -> LazyValue:
        if self.default_value is not None:
            raise DupOperationException("Default value already defined")
        self.default_value = value
        return self

    def converter(self, func: Callable[[str], Any]) -> LazyValue:
        if self.converter_func is not None:
            raise DupOperationException("Converter already defined")
        self.converter_func = func
        return self

    def optional(self) -> LazyValue:
        if self.is_optional is True:
            raise DupOperationException("Value already optional")
        self.is_optional = True
        return self

    def validator(self, validator_func: Callable[[str], None]) -> LazyValue:
        if self.validator_func is not None:
            raise DupOperationException("Validator already defined")
        self.validator_func = validator_func
        return self

    def resolve(self) -> Any:
        value = self.get_raw_value()
        if value is None:
            return None

        if self.validator_func is not None:
            self.validator_func(value)

        if self.converter_func is not None:
            return self.converter_func(value)
        return value

    def get_raw_value(self) -> str | None:
        value = self.environment._get(self.key)
        if value is None:
            value = self.default_value
        if value is None and not self.is_optional:
            raise NoValueException(f"Value for key {self.key} does not exists")
        return value



from __future__ import annotations

import os
from typing import Any

from . import exceptions
from .imanager import IEnvironmentManager
from .immutable_dict import ImmutableDict
from .lazy_value import LazyValue


class EnvironmentManager(IEnvironmentManager):
    exc = exceptions
    _mapping: ImmutableDict[str, str]
    _init: bool = False

    def __init__(self, mapping: dict[str, str]) -> None:
        self._mapping = ImmutableDict(mapping)
        self._init = True

    def get(self, key: str) -> LazyValue:
        return LazyValue(key, self)

    def _get(self, key: str) -> str | None:
        return self._mapping.get(key)

    def __setattr__(self, key: str, value: Any) -> None:
        if key == "_mapping" and self._init:
            raise AttributeError("_mapping is readonly attribute")
        if key == "_init" and self._init:
            raise AttributeError("_init is readonly attribute")
        return super().__setattr__(key, value)

    def __delattr__(self, key: str) -> None:
        if key == "_mapping":
            raise AttributeError("_mapping is readonly attribute")
        if key == "_init":
            raise AttributeError("_init is readonly attribute")
        return super().__delattr__(key)

    @classmethod
    def load(cls) -> EnvironmentManager:
        return cls(dict(os.environ))



import resource


def get_resource_limit(n: int) -> tuple[int | None, int | None]:
    r = resource.getrlimit(n)
    return (
        r[0] if r[0] != resource.RLIM_INFINITY else None,
        r[1] if r[1] != resource.RLIM_INFINITY else None,
    )


def parse_dns_servers_from_file(
    filename: str = "/etc/resolv.conf",
) -> list[str]:
    with open(filename, "r") as file:
        return list(
            map(
                # Line example "nameserver 192.168.1.1"
                lambda line: line.replace("nameserver ", ""),
                filter(
                    lambda line: line.startswith("nameserver"),
                    file.read().splitlines(),
                ),
            )
        )



from __future__ import annotations

import platform
from dataclasses import dataclass


@dataclass(frozen=True)
class OSInfo:
    system: str
    release: str
    machine: str
    version: str

    @classmethod
    def load(cls) -> OSInfo:
        return cls(
            machine=platform.machine(),
            release=platform.release(),
            system=platform.system(),
            version=platform.version(),
        )



from __future__ import annotations

import platform
from dataclasses import dataclass


@dataclass(frozen=True)
class PythonInfo:
    version: str
    build_date: str
    implementation: str

    @classmethod
    def load(cls) -> PythonInfo:
        _, build_date = platform.python_build()
        return cls(
            build_date=build_date,
            version=platform.python_version(),
            implementation=platform.python_implementation(),
        )



from abc import ABC, abstractmethod

from .execution_context import ExecutionContext


class IMicroService(ABC):
    @abstractmethod
    def __init__(self, execution_context: ExecutionContext) -> None:
        raise NotImplementedError

    @abstractmethod
    async def setup(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def main(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def shutdown(self) -> None:
        raise NotImplementedError



from typing import Any, TypeVar

KeyType = TypeVar("KeyType")
ValueType = TypeVar("ValueType")


class ImmutableDict(dict[KeyType, ValueType]):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    def __setitem__(self, *args: Any, **kwargs: Any) -> None:
        raise AttributeError("Setting item is prohibited")

    def __delitem__(self, *args: Any, **kwargs: Any) -> None:
        raise AttributeError("Deleting item is prohibited")



from __future__ import annotations

import resource
from dataclasses import dataclass

from .utils import get_resource_limit


@dataclass(frozen=True)
class Limit:
    soft: int | None
    hard: int | None


@dataclass(frozen=True)
class ResourceLimits:
    page_size: int
    max_cpu_time: Limit
    max_file_size: Limit
    max_ram_heap: Limit
    max_ram_stack: Limit
    max_open_files: Limit
    max_child_processes: Limit

    @classmethod
    def load(cls) -> ResourceLimits:
        return ResourceLimits(
            page_size=resource.getpagesize(),
            max_cpu_time=Limit(*get_resource_limit(resource.RLIMIT_CPU)),
            max_file_size=Limit(*get_resource_limit(resource.RLIMIT_FSIZE)),
            max_ram_heap=Limit(*get_resource_limit(resource.RLIMIT_DATA)),
            max_ram_stack=Limit(*get_resource_limit(resource.RLIMIT_STACK)),
            max_child_processes=Limit(
                *get_resource_limit(resource.RLIMIT_NPROC)
            ),
            max_open_files=Limit(*get_resource_limit(resource.RLIMIT_NOFILE)),
        )



import os
from pathlib import Path

from setuptools import setup  # type: ignore [import]


setup(
    name="mcrs",
    version=os.environ["GITHUB_REF_NAME"],
    description="Package for microservices",
    author="Vladimir Vojtenko",
    author_email="vladimirdev635@gmail.com",
    license="MIT",
    packages=["mcrs", "mcrs.lifespan", "mcrs.environment"],
    package_data={
        "mcrs": ["py.typed"],
        "mcrs.lifespan": ["py.typed"],
        "mcrs.environment": ["py.typed"],
    },
    long_description=(Path(__file__).parent / "README.md").read_text(),
    long_description_content_type="text/markdown",
)



import time
from multiprocessing import Event, Process
from multiprocessing.synchronize import Event as EventClass

from .delayed_interrupt import DelayedInterrupt


def process_main(
    terminate_event: EventClass, result_event: EventClass
) -> None:
    with DelayedInterrupt():
        terminate_event.set()
        time.sleep(1)
        result_event.set()


def test_delayed_interrupt() -> None:
    terminate_event = Event()
    result_event = Event()
    process = Process(
        target=process_main, args=(terminate_event, result_event)
    )
    process.start()
    terminate_event.wait(timeout=10)
    process.terminate()
    process.join()
    assert result_event.is_set()



from .config import EnvConfValue, EnvironmentConfig
from .exceptions import (
    DupOperationException,
    EnvironmentException,
    LazyValueException,
    NoValueException,
)
from .imanager import IEnvironmentManager
from .immutable_dict import ImmutableDict
from .lazy_value import LazyValue
from .manager import EnvironmentManager

__all__ = (
    "EnvironmentManager",
    "ImmutableDict",
    "EnvironmentException",
    "NoValueException",
    "LazyValueException",
    "DupOperationException",
    "LazyValue",
    "EnvConfValue",
    "EnvironmentConfig",
)
